import os
import re
from typing import Mapping
from jinja2 import Template
from .log_support import logger, sql_id_log, page_sql_id_log
from .support import get_named_sql_args, SqlModel, MapperError, is_dynamic_sql, get_batch_args, SqlAction
from .db import do_get, do_query, do_query_one, do_execute, do_select_page, do_select, do_select_one, do_query_page, init_db as _init_db, \
    get_connection as _get_connection, insert as _insert, save as _save, batch_insert as _batch_insert, batch_execute as _batch_execute

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

_SQL_CONTAINER = dict()
_valid_sql_actions = (SqlAction.INSERT.value, SqlAction.UPDATE.value, SqlAction.DELETE.value, SqlAction.SELECT.value)


def init_db(user='root', password='', database='test', host='127.0.0.1', port=3306, pool_size=5, use_unicode=True, show_sql=False,
        mapper_path='mapper', **kwargs):
    load_mapper(mapper_path)
    _init_db(user, password, database, host, port, pool_size, use_unicode, show_sql, **kwargs)


def insert(table: str, **kwargs):
    """
    Insert data into table, return effect rowcount.
    :param table: table name
    :param kwargs: {'name': '张三', 'age': 20}
    return: Effect rowcount
    """
    return _insert(table, **kwargs)


def save(table: str, **kwargs):
    """
    Insert data into table, return primary key.
    :param table: table name
    :param kwargs: {'name': '张三', 'age': 20}
    :return: Primary key
    """
    return _save(table, **kwargs)


def execute(sql_id: str, *args, **kwargs):
    """
    Execute SQL.
    sql: INSERT INTO user(name, age) VALUES(?, ?)  -->  args: ('张三', 20)
         INSERT INTO user(name, age) VALUES(:name,:age)  -->  kwargs: {'name': '张三', 'age': 20}
    """
    sql_id_log('dbx.execute', sql_id, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_execute(sql, *args)


def batch_insert(table: str, *args):
    """
    Batch insert
    :param table: table name
    :param args: All number must have same key. [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
    :return: Effect row count
    """
    return _batch_insert(table, *args)


def batch_execute(sql_id: str, *args):
    """
    Batch execute
    sql: INSERT INTO user(name, age) VALUES(?, ?)  -->  args: [('张三', 20), ('李四', 28)]
         INSERT INTO user(name, age) VALUES(:name,:age)  -->  args: [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
    :return: Effect row count
    """
    logger.debug("Exec func 'mysqlx.dbx.%s' \n\t\t sql_id: '%s' \n\t\t args: %s" % ('batch_execute', sql_id, args))
    args = get_batch_args(*args)
    sql, _ = do_get_sql(get_sql_model(sql_id), True, None, *args)
    return _batch_execute(sql, *args)


# ----------------------------------------------------------Query function------------------------------------------------------------------
def get(sql_id: str, *args, **kwargs):
    """
    Execute select SQL and expected one int and only one int result. Automatically add 'limit ?' behind the sql statement if not.
    MultiColumnsError: Expect only one column.
    sql: SELECT count(1) FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
         SELECT count(1) FROM user WHERE name=:name and age=:age limit 1  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_id_log('get', sql_id, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_get(sql, *args)


def query(sql_id: str, *args, **kwargs):
    """
    Execute select SQL and return list or empty list if no result.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_id_log('query', sql_id, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_query(sql, *args)


def query_one(sql_id: str, *args, **kwargs):
    """
    Execute select SQL and expected one row result(dict). Automatically add 'limit ?' behind the sql statement if not.
    If no result found, return None.
    If multiple results found, the first one returned.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1 -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age limit 1  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_id_log('query_one', sql_id, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_query_one(sql, *args)


def select(sql_id: str, *args, **kwargs):
    """
    Execute select SQL and return list(tuple) or empty list if no result.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age   -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_id_log('select', sql_id, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_select(sql, *args)


def select_one(sql_id: str, *args, **kwargs):
    """
    Execute select SQL and expected one row result(tuple). Automatically add 'limit ?' behind the sql statement if not.
    If no result found, return None.
    If multiple results found, the first one returned.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age limit 1  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_id_log('select_one', sql_id, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_select_one(sql, *args)


def query_page(sql_id: str, page_num=1, page_size=10, *args, **kwargs):
    """
    Execute select SQL and return list or empty list if no result. Automatically add 'limit ?,?' after sql statement if not.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    page_sql_id_log('query_page', sql_id, page_num, page_size, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_query_page(sql, page_num, page_size, *args)


def select_page(sql_id: str, page_num=1, page_size=10, *args, **kwargs):
    """
    Execute select SQL and return list or empty list if no result. Automatically add 'limit ?,?' after sql statement if not.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    page_sql_id_log('select_page', sql_id, page_num, page_size, *args, **kwargs)
    sql, args = get_sql(sql_id, *args, **kwargs)
    return do_select_page(sql, page_num, page_size, *args)


def get_connection():
    return _get_connection()


def get_sql(sql_id: str, *args, **kwargs):
    sql_model = get_sql_model(sql_id)
    return do_get_sql(sql_model, False, None, *args, **kwargs)


def do_get_sql(sql_model, batch, param_names, *args, **kwargs):
    """
    Get sql from SqlModel.
    :param sql_model: SqlModel
    :param batch: bool, is batch or not
    :param param_names: original function parameter names
    :param args:
    :param kwargs:
    :return:
    """
    if sql_model.dynamic:
        if not kwargs:
            raise MapperError("Parameter 'kwargs' must not be empty when named mapping sql.")
        sql = sql_model.sql.render(**kwargs)
        logger.debug("Original sql: {}".format(sql))
        return get_named_sql_args(sql, **kwargs)
    else:
        logger.debug("Original sql: {}".format(sql_model.sql))
        if sql_model.mapping and kwargs:
            return get_named_sql_args(sql_model.sql, **kwargs)
        elif sql_model.placeholder and kwargs:
            logger.warning("Better use 'func(arg1, arg2...)' then 'func(arg1=arg1, arg2=arg2...)' if sql contain '?' placeholder.")
            args = [kwargs[name] for name in param_names if name in kwargs] if param_names else list(kwargs.values())
        elif sql_model.mapping and not kwargs and (not batch or
                                                   (batch and (not args or not isinstance(args[0], Mapping)))):  # batch_execute时args可能为List[Mapping]
            raise MapperError("Parameter 'kwargs' must not be empty when named mapping sql.")
        return sql_model.sql, args


def build_sql_id(namespace, _id):
    return namespace + "." + _id


def get_sql_model(sql_id: str):
    global _SQL_CONTAINER
    return _SQL_CONTAINER[sql_id]


# ----------------------------------------------------------Load mapper--------------------------------------------------------------------
def load_mapper(path: str):
    if os.path.isfile(path) and path.endswith(".xml"):
        _parse_mapper_file(path)
    elif os.path.isdir(path):
        for f in os.listdir(path):
            file = os.path.join(path, f)
            if os.path.isfile(file) and f.endswith(".xml"):
                _parse_mapper_file(file)
            elif os.path.isdir(file):
                load_mapper(file)


def _parse_mapper_file(file: str):
    global _SQL_CONTAINER
    tree = ET.parse(file)
    root = tree.getroot()
    namespace = root.attrib.get('namespace', '')
    results = list(map(lambda child: _load_sql(namespace, child, file), root))
    sql_ids, file_all_includes = zip(*results)
    for i, includes in enumerate(file_all_includes):
        if includes:
            for include in includes:
                if include not in sql_ids:
                    raise MapperError("Include '%s' are not exist in mapper file: %s" % (include, file))

                include_sql_id = build_sql_id(namespace, include)
                include_sql_model = _SQL_CONTAINER[include_sql_id]
                if include_sql_model.includes:
                    raise MapperError("Nested include: '%s' include '%s' and it include %s in mapper file: %s" % (
                        sql_ids[i], include, include_sql_model.includes, file))

    # include sql
    include_results = filter(lambda x: x[1] is not None, results)
    for sql_id, includes in include_results:
        _handle_includes(build_sql_id(namespace, sql_id), includes)

    # dynamic sql change to Template
    for sql_id in sql_ids:
        sql_model = _SQL_CONTAINER[build_sql_id(namespace, sql_id)]
        if sql_model.dynamic:
            sql_model.sql = Template(sql_model.sql)


def _handle_includes(sql_id, includes):
    is_dynamic = False
    sql_model = _SQL_CONTAINER[sql_id]
    for include in includes:
        include_sql_id = build_sql_id(sql_model.namespace, include)
        include_sql_model = _SQL_CONTAINER[include_sql_id]
        if include_sql_model.dynamic:
            is_dynamic = True
        if include_sql_model.includes:
            _handle_includes(include_sql_id, include_sql_model.includes)
        sql = re.sub(r'{{\s*%s\s*}}' % include, include_sql_model.sql, sql_model.sql)

    _valid_sql(sql_id, sql, sql_model.action)
    if is_dynamic or is_dynamic_sql(sql):
        sql_model.dynamic = True
        sql_model.mapping = True
        sql_model.placeholder = False
    else:
        sql_model.mapping = ':' in sql
        sql_model.placeholder = False if sql_model.mapping else '?' in sql
    sql_model.sql = sql
    sql_model.includes = None


def _load_sql(namespace, child, file):
    global _SQL_CONTAINER
    includes = None
    _id = child.attrib.get('id')
    assert _id, "Mapper 'id' must be set in mapper file: %s." % file
    sql_id = build_sql_id(namespace, _id)
    assert sql_id not in _SQL_CONTAINER, "Sql id '%s' repeat." % sql_id
    include = child.attrib.get('include')
    sql = child.text.strip()
    if include:
        includes = include.split(",")
        for include in set(includes):
            assert include != _id, "Include must not be it self, id: '%s' = include: '%s' " % (_id, include)
        _SQL_CONTAINER[sql_id] = SqlModel(sql=sql, action=child.tag, namespace=namespace, includes=includes)
    elif is_dynamic_sql(sql):
        _valid_sql(sql_id, sql, child.tag)
        _SQL_CONTAINER[sql_id] = SqlModel(sql=sql, action=child.tag, namespace=namespace, dynamic=True)
    else:
        _valid_sql(sql_id, sql, child.tag)
        _SQL_CONTAINER[sql_id] = SqlModel(sql=sql, action=child.tag, namespace=namespace)

    return _id, includes


def _valid_sql(sql_id, sql, tag):
    assert tag in _valid_sql_actions and tag in sql.lower(), "Sql id: '{}' has not '{}' key word in {} sql.".format(sql_id, tag, tag)



import time
from .support import DB_LOCK

_SNOWFLAKE = None
_WORKER_BITS = 10
_SEQUENCE_BITS = 12
_EPOCH = 1688140800000
# _SEQUENCE_MASK = 4095


def init_snowflake(machine_id=1, epoch=_EPOCH, worker_bits=_WORKER_BITS, sequence_bits=_SEQUENCE_BITS):
    global _SNOWFLAKE
    _SNOWFLAKE = Snowflake(machine_id, epoch, worker_bits, sequence_bits)


def get_id():
    global _SNOWFLAKE
    try:
        return _SNOWFLAKE.generate_id()
    except AttributeError:
        raise RuntimeError("Please init Snowflake first with: snowflake.init_snowflake(machine_id: int, epoch: int, worker_bits: int, sequence_bits: int)")


def _get_timestamp():
    return int(time.time() * 1000)


class Snowflake:
    def __init__(self, machine_id: int, epoch: int, worker_bits: int, sequence_bits: int):
        self.machine_id = machine_id
        self.epoch = epoch
        self.sequence = 0
        self.last_timestamp = -1
        self.worker_shift = sequence_bits
        self.sequence_mask = -1 ^ (-1 << sequence_bits)
        self.timestamp_left_shift = worker_bits + sequence_bits

        maxWorkerId = -1 ^ (-1 << worker_bits);
        assert 0 < machine_id <= maxWorkerId, 'machine_id must ge 0 le %d' % maxWorkerId
        assert 10 <= self.timestamp_left_shift <= 22, 'worker_bits add sequence_bits must between 10 and 20, but it is %d' % self.timestamp_left_shift

    def generate_id(self):
        with DB_LOCK:
            timestamp = _get_timestamp()
            if timestamp < self.last_timestamp:
                raise Exception("Clock moved backwards")
            if timestamp == self.last_timestamp:
                self.sequence = (self.sequence + 1) & self.sequence_mask
                if self.sequence == 0:
                    timestamp = self._wait_next_millis(self.last_timestamp)
            else:
                self.sequence = 0
            self.last_timestamp = timestamp
            return ((timestamp - self.epoch) << self.timestamp_left_shift) | (self.machine_id << self.worker_shift) | self.sequence
            # return ((timestamp - 1288834974657) << 22) | (self.machine_id << 12) | self.sequence

    @staticmethod
    def _wait_next_millis(last_timestamp):
        timestamp = _get_timestamp()
        while timestamp <= last_timestamp:
            timestamp = _get_timestamp()
        return timestamp




import re
import threading
from enum import Enum
from jinja2 import Template
from functools import lru_cache
from .log_support import logger
from typing import Sequence, List
from .constant import NAMED_REGEX, DYNAMIC_REGEX

DB_LOCK = threading.RLock()


def get_batch_args(*args):
    return args[0] if isinstance(args, tuple) and len(args) == 1 and isinstance(args[0], Sequence) else args


def try_commit(db_ctx):
    if db_ctx.transactions == 0:
        logger.debug('Commit transaction...')
        try:
            db_ctx.connection.commit()
            logger.debug('Commit ok.')
        except Exception:
            logger.warning('Commit failed, try rollback...')
            db_ctx.connection.rollback()
            logger.warning('Rollback ok.')
            raise


def simple_sql(sql: str, *args, **kwargs):
    return get_named_sql_args(sql, **kwargs) if kwargs else (sql, args)


def dynamic_sql(sql: str, *args, **kwargs):
    sql_type = _get_sql_type(sql)
    if sql_type >= 1 and not kwargs:
        raise MapperError("Parameter 'kwargs' must not be empty when named mapping sql.")
    if sql_type == 0:
        return sql, args
    if sql_type == 1:
        sql = Template(sql).render(**kwargs)
    return get_named_sql_args(sql, **kwargs)


def is_dynamic_sql(sql: str):
    return re.search(DYNAMIC_REGEX, sql)


@lru_cache(maxsize=256)
def _get_sql_type(sql: str):
    """
    :return: 0: placeholder, 1: dynamic, 2: named mapping
    """
    if is_dynamic_sql(sql):
        return 1
    if ':' in sql:
        return 2
    return 0


def get_named_sql_args(sql: str, **kwargs):
    args = get_named_args(sql, **kwargs)
    return get_named_sql(sql), args


@lru_cache(maxsize=256)
def get_named_sql(sql: str):
    return re.sub(NAMED_REGEX, '?', sql)


def get_named_args(sql: str, **kwargs):
    return [kwargs[r[1:]] for r in re.findall(NAMED_REGEX, sql)]


class DBCtx(threading.local):
    """
    Thread local object that holds connection info.
    """

    def __init__(self, connect, use_mysql_connector):
        self.connect = connect
        self.connection = None
        self.transactions = 0
        self.prepared = True
        if use_mysql_connector:
            self.get_cursor = lambda: self.connection.cursor(prepared=self.prepared)
            self.log = lambda action: logger.debug('%s connection <%s>...' % (action, hex(id(self.connection._cnx))))
        else:
            self.get_cursor = lambda: self.connection.cursor()
            self.log = lambda action: logger.debug('%s connection <%s>...' % (action, hex(id(self.connection))))

    def is_not_init(self):
        return self.connection is None

    def init(self):
        self.transactions = 0
        self.connection = self.connect()
        self.log('Use')

    def release(self):
        if self.connection:
            self.log('Release')
            self.connection.close()
            self.connection = None

    def cursor(self):
        """
        Return cursor
        """
        # logger.debug('Cursor prepared: %s' % self.prepared)
        return self.get_cursor()

    def statement(self, sql: str):
        """
        Return statement
        """
        return self.connection.statement(sql)


class ConnectionCtx(object):
    """
    ConnectionCtx object that can open and close connection context. ConnectionCtx object can be nested and only the most
    outer connection has effect.
    with connection():
        pass
        with connection():
            pass
    """

    def __init__(self, db_ctx):
        self.db_ctx = db_ctx

    def __enter__(self):
        self.should_cleanup = False
        if self.db_ctx.is_not_init():
            self.db_ctx.init()
            self.should_cleanup = True
        return self

    def __exit__(self, exctype, excvalue, traceback):
        if self.should_cleanup:
            self.db_ctx.release()


class TransactionCtx(object):
    """
    TransactionCtx object that can handle transactions.
    with TransactionCtx():
        pass
    """

    def __init__(self, db_ctx):
        self.db_ctx = db_ctx

    def __enter__(self):
        self.should_close_conn = False
        if self.db_ctx.is_not_init():
            # needs open a connection first:
            self.db_ctx.init()
            self.should_close_conn = True
        self.db_ctx.transactions += 1
        logger.debug('Begin transaction...' if self.db_ctx.transactions == 1 else 'Join current transaction...')
        return self

    def __exit__(self, exctype, excvalue, traceback):
        self.db_ctx.transactions -= 1
        try:
            if self.db_ctx.transactions == 0:
                if exctype is None:
                    self.commit()
                else:
                    self.rollback()
        finally:
            if self.should_close_conn:
                self.db_ctx.release()

    def commit(self):
        try_commit(self.db_ctx)

    def rollback(self):
        logger.warning('Rollback transaction...')
        self.db_ctx.connection.rollback()
        logger.debug('Rollback ok.')


class DBError(Exception):
    pass


class MapperError(DBError):
    pass


class MultiColumnsError(DBError):
    pass


class Dict(dict):
    def __init__(self, names=(), values=(), **kw):
        super(Dict, self).__init__(**kw)
        for k, v in zip(names, values):
            self[k] = v

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Dict' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        self[key] = value


class SqlModel:
    def __init__(self, sql: str, action: str, namespace: str, dynamic=False, includes: List[str] = None):
        self.sql = sql
        self.action = action
        self.namespace = namespace
        self.dynamic = dynamic
        self.includes = includes
        self.mapping = True if dynamic else ':' in sql
        self.placeholder = False if self.mapping else '?' in sql


class SqlAction(Enum):
    CALL = 'call'
    INSERT = 'insert'
    UPDATE = 'update'
    DELETE = 'delete'
    SELECT = 'select'



import os
import functools
from .log_support import logger
from .dbx import get_sql_model, do_get_sql, build_sql_id
from .support import SqlAction, get_named_sql_args, simple_sql, MapperError
from .db import do_get, do_query, do_query_one, do_execute, do_batch_execute, do_select, do_select_one, batch_execute, do_save

_UPDATE_ACTIONS = (SqlAction.INSERT.value, SqlAction.UPDATE.value, SqlAction.DELETE.value, SqlAction.CALL.value)


def mapper(namespace: str = None, sql_id: str = None, batch=False, return_pk=False):
    def _decorator(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            param_names = func.__code__.co_varnames
            full_sql_id, func_name = _before(func, namespace, sql_id, *args, **kwargs)
            sql_model = get_sql_model(full_sql_id)
            exec_func = _get_exec_func(func, sql_model.action, batch, return_pk)
            if not batch:
                use_sql, args = do_get_sql(sql_model, batch, param_names, *args, **kwargs)
                if return_pk and SqlAction.INSERT.value not in use_sql.lower():
                    raise MapperError("Only insert sql can return primary key.")
            else:
                if kwargs:
                    logger.warning("Batch exec sql better use like '{}(args)' or '{}(*args)' then '{}(args=args)'".format(func_name, func_name, func_name))
                    args = list(kwargs.values())[0]
                use_sql, _ = do_get_sql(sql_model, batch, param_names, *args)
            return exec_func(use_sql, *args)

        return _wrapper
    return _decorator


def sql(value: str, batch=False, return_pk=False):
    def _decorator(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            use_sql = value
            low_sql = value.lower()
            if any([action in low_sql for action in _UPDATE_ACTIONS]):
                if batch:
                    if kwargs:
                        args = list(kwargs.values())[0]
                    return batch_execute(use_sql, *args)
                if kwargs:
                    use_sql, args = get_named_sql_args(**kwargs)
                if return_pk:
                    assert SqlAction.INSERT.value in use_sql.lower(), 'Only insert sql can return primary key.'
                    return do_save(use_sql, *args)
                return do_execute(use_sql, *args)
            elif SqlAction.SELECT.value in low_sql:
                select_func = _get_select_func(func)
                use_sql, args = simple_sql(use_sql, *args, **kwargs)
                return select_func(use_sql, *args)
            else:
                return ValueError("Invalid sql: {}.".format(sql))

        return _wrapper
    return _decorator


def _get_exec_func(func, action, batch, return_pk):
    if action == SqlAction.SELECT.value:
        return _get_select_func(func)
    elif batch:
        return do_batch_execute
    elif return_pk:
        return do_save
    else:
        return do_execute


def _get_select_func(func):
    names = func.__code__.co_names
    is_list = 'list' in names or 'List' in names
    if 'Mapping' in names and is_list:
        return do_query
    elif 'Mapping' in names:
        return do_query_one
    elif len(names) == 1 and names[0] in ('int', 'float', 'Decimal', 'str', 'AnyStr', 'date', 'time', 'datetime'):
        return do_get
    elif len(names) == 1 and names[0] in ('tuple', 'Tuple'):
        return do_select_one
    elif is_list:
        return do_select
    else:
        return do_query


def _before(func, namespace, _id, *args, **kwargs):
    file_name = os.path.basename(func.__code__.co_filename)[:-3]
    _namespace = namespace if namespace else file_name
    _id = _id if _id else func.__name__
    sql_id = build_sql_id(_namespace, _id)
    func_name = file_name + '.' + func.__name__
    logger.debug("Exec mapper func: '%s', sql_id: '%s', args: %s, kwargs: %s" % (func_name, sql_id, args, kwargs))
    return sql_id, func_name



import functools
from typing import Sequence
from .constant import LIMIT_1, MAPPER_PATH, PK_SQL
from .log_support import logger, sql_log, page_log, insert_log, do_sql_log, do_page_log
from .support import DBCtx, ConnectionCtx, Dict, MultiColumnsError, TransactionCtx, try_commit, dynamic_sql, get_named_sql, get_named_args, DBError, \
    DB_LOCK, get_batch_args

_DB_CTX = None
_SHOW_SQL = False


def init_db(user='root', password='', database='test', host='127.0.0.1', port=3306, pool_size=5, use_unicode=True, show_sql=False, **kwargs):
    global _DB_CTX
    global _SHOW_SQL
    use_mysql_connector = False
    try:
        from mysql.connector import connect
        use_mysql_connector = True
    except ImportError:
        try:
            from pymysql import connect
        except ImportError:
            raise DBError("Please install mysql-connector-python or PyMySQL, and mysql-connector-python >= 8.0.13, PyMySQL >= 0.9.0.")

    if 'debug' in kwargs:
        if kwargs['debug']:
            from logging import DEBUG
            logger.setLevel(DEBUG)
        del kwargs['debug']

    is_pool = use_mysql_connector and pool_size >= 1
    with DB_LOCK:
        if _DB_CTX is not None:
            raise DBError('DB is already initialized.')
        _SHOW_SQL = show_sql
        if MAPPER_PATH in kwargs:
            from .dbx import load_mapper
            load_mapper(kwargs[MAPPER_PATH])
            del kwargs[MAPPER_PATH]
        if is_pool:
            kwargs['pool_size'] = pool_size
            if 'pool_name' not in kwargs:
                kwargs['pool_name'] = "{}_pool".format(database)
        kwargs['user'] = user
        kwargs['password'] = password
        kwargs['database'] = database
        kwargs['host'] = host
        kwargs['port'] = port
        kwargs['use_unicode'] = use_unicode
        _DB_CTX = DBCtx(connect=lambda: connect(**kwargs), use_mysql_connector=use_mysql_connector)
    if is_pool:
        logger.info('Init db engine <%s> ok with connection pool size: %d.' % (hex(id(_DB_CTX)), pool_size))
    else:
        logger.info('Init db engine <%s> ok without connection pool.' % hex(id(_DB_CTX)))


def connection():
    """
    Return _ConnectionCtx object that can be used by 'with' statement:
    with connection():
        pass
    """
    global _DB_CTX
    return ConnectionCtx(_DB_CTX)


def with_connection(func):
    """
    Decorator for reuse connection.
    @with_connection
    def foo(*args, **kw):
        f1()
        f2()
    """

    global _DB_CTX

    @functools.wraps(func)
    def _wrapper(*args, **kw):
        with ConnectionCtx(_DB_CTX):
            return func(*args, **kw)

    return _wrapper


def transaction():
    """
    Create a transaction object so can use with statement:
    with transaction():
        pass
    with transaction():
         insert(...)
         update(... )
    """
    global _DB_CTX
    return TransactionCtx(_DB_CTX)


def with_transaction(func):
    """
    A decorator that makes function around transaction.
    @with_transaction
    def update_profile(id, name, rollback):
         u = dict(id=id, name=name, email='%s@test.org' % name, passwd=name, last_modified=time.time())
         insert('user', **u)
         r = update('update user set passwd=? where id=?', name.upper(), id)
    """
    global _DB_CTX

    @functools.wraps(func)
    def _wrapper(*args, **kw):
        with TransactionCtx(_DB_CTX):
            return func(*args, **kw)

    return _wrapper


# ----------------------------------------------------------Update function------------------------------------------------------------------
def insert(table: str, **kwargs):
    """
    Insert data into table, return effect rowcount.
    :param table: table name
    :param kwargs: name='张三', age=20}
    return: Effect rowcount
    """
    insert_log('insert', table, **kwargs)
    sql, args = _insert_sql_args(table.strip(), **kwargs)
    return do_execute(sql, *args)


def save(table: str, **kwargs):
    """
    Insert data into table, return primary key.
    :param table: table
    :param kwargs: name='张三', age=20
    :return: Primary key
    """
    insert_log('save', table, **kwargs)
    sql, args = _insert_sql_args(table.strip(), **kwargs)
    return do_save(sql, *args)


@with_connection
def do_save(sql: str, *args):
    """
    Insert data into table, return primary key.
    :param sql: table
    :param args:
    :return: Primary key
    """
    global _DB_CTX
    cursor = None
    sql = _before_execute('do_save', sql, *args)
    try:
        cursor = _DB_CTX.connection.cursor()
        cursor.execute(sql, args)
        cursor.execute(PK_SQL)
        result = cursor.fetchone()
        try_commit(_DB_CTX)
        return result[0]
    finally:
        if cursor:
            cursor.close()


def execute(sql: str, *args, **kwargs):
    """
    Execute SQL.
    sql: INSERT INTO user(name, age) VALUES(?, ?)  -->  args: ('张三', 20)
         INSERT INTO user(name, age) VALUES(:name,:age)  -->  kwargs: {'name': '张三', 'age': 20}
    """
    sql_log('db.execute', sql, *args, **kwargs)
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_execute(sql, *args)


def batch_insert(table: str, *args):
    """
    Batch insert
    :param table: table name
    :param args: All number must have same key. [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
    :return: Effect row count
    """
    logger.debug("Exec func 'mysqlx.db.%s' \n\t\t Table: '%s', args: %s" % ('batch_insert', table, args))
    assert len(args) > 0, 'args must not be empty.'
    sql, args = _batch_insert_sql_args(table, *args)
    return batch_execute(sql, *args)


def batch_execute(sql: str, *args):
    """
    Batch execute
    :param sql: INSERT INTO user(name, age) VALUES(?, ?)  -->  args: [('张三', 20), ('李四', 28)]
                INSERT INTO user(name, age) VALUES(:name,:age)  -->  args: [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
    :param args: All number must have same size.
    :return: Effect row count
    """
    logger.debug("Exec func 'mysqlx.db.%s' \n\t\t sql: '%s' \n\t\t args: %s" % ('batch_execute', sql, args))
    assert len(args) > 0, 'args must not be empty.'
    if isinstance(args[0], dict):
        sql, args = _batch_named_sql_args(sql, *args)

    return do_batch_execute(sql, *args)


# ----------------------------------------------------------Query function------------------------------------------------------------------
def get(sql: str, *args, **kwargs):
    """
    Execute select SQL and expected one int and only one int result. Automatically add 'limit ?' after sql statement if not.
    MultiColumnsError: Expect only one column.
    sql: SELECT count(1) FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
         SELECT count(1) FROM user WHERE name=:name and age=:age limit 1  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_log('get', sql, *args, **kwargs)
    global _DB_CTX
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_get(sql, *args)


def query(sql: str, *args, **kwargs):
    """
    Execute select SQL and return list or empty list if no result.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_log('query', sql, *args, **kwargs)
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_query(sql, *args)


def query_one(sql: str, *args, **kwargs):
    """
    Execute select SQL and expected one row result(dict). Automatically add 'limit ?' after sql statement if not.
    If no result found, return None.
    If multiple results found, the first one returned.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1 -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age limit 1  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_log('query_one', sql, *args, **kwargs)
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_query_one(sql, *args)


def select(sql: str, *args, **kwargs):
    """
    Execute select SQL and return list(tuple) or empty list if no result.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age   -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_log('select', sql, *args, **kwargs)
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_select(sql, *args)


def select_one(sql: str, *args, **kwargs):
    """
    Execute select SQL and expected one row result(tuple). Automatically add 'limit ?' after sql statement if not.
    If no result found, return None.
    If multiple results found, the first one returned.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age limit 1  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    sql_log('select_one', sql, *args, **kwargs)
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_select_one(sql, *args)


def query_page(sql: str, page_num=1, page_size=10, *args, **kwargs):
    """
    Execute select SQL and return list or empty list if no result. Automatically add 'limit ?,?' after sql statement if not.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age  -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    page_log('query_page', sql, page_num, page_size, *args, **kwargs)
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_query_page(sql, page_num, page_size, *args)


def select_page(sql: str, page_num=1, page_size=10, *args, **kwargs):
    """
    Execute select SQL and return list(tuple) or empty list if no result. Automatically add 'limit ?,?' after sql statement if not.
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
         SELECT * FROM user WHERE name=:name and age=:age   -->  kwargs: ('张三', 20) --> kwargs: {'name': '张三', 'age': 20}
    """
    page_log('select_page', sql, page_num, page_size, *args, **kwargs)
    sql, args = dynamic_sql(sql, *args, **kwargs)
    return do_select_page(sql, page_num, page_size, *args)


# ----------------------------------------------------------Do function------------------------------------------------------------------
@with_connection
def do_execute(sql: str, *args):
    """
    Execute sql return effect rowcount
    sql: insert into user(name, age) values(?, ?)  -->  args: ('张三', 20)
    """
    global _DB_CTX
    cursor = None
    sql = _before_execute('do_execute', sql.strip(), *args)
    try:
        cursor = _DB_CTX.connection.cursor()
        cursor.execute(sql, args)
        effect_rowcount = cursor.rowcount
        try_commit(_DB_CTX)
        return effect_rowcount
    finally:
        if cursor:
            cursor.close()


@with_connection
def do_batch_execute(sql: str, *args):
    """
    Batch execute sql return effect rowcount
    :param sql: insert into user(name, age) values(?, ?)  -->  args: [('张三', 20), ('李四', 28)]
    :param args: All number must have same size.
    :return: Effect rowcount
    """
    global _DB_CTX
    cursor = None
    sql = _before_execute('do_batch_execute', sql.strip(), *args)
    args = get_batch_args(*args)
    try:
        cursor = _DB_CTX.cursor()
        cursor.executemany(sql, args)
        effect_rowcount = cursor.rowcount
        try_commit(_DB_CTX)
        return effect_rowcount
    finally:
        if cursor:
            cursor.close()


def do_get(sql: str, *args):
    """
    Execute select SQL and expected one int and only one int result. Automatically add 'limit ?' behind the sql statement if not.
    MultiColumnsError: Expect only one column.
    sql: SELECT count(1) FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
    """
    do_sql_log('do_get', sql, *args)
    sql, args = _limit_one_sql_args(sql, *args)
    return do_get_limit(sql, *args)


def do_get_limit(sql: str, *args):
    """
    Execute select SQL and expected one int and only one int result, SQL contain 'limit'.
    MultiColumnsError: Expect only one column.
    sql: SELECT count(1) FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
    """
    do_sql_log('do_get_with_limit', sql, *args)
    result = do_select_one_limit(sql, *args)
    if result:
        if len(result) == 1:
            return result[0]
        msg = "Exec func 'mysqlx.db.%s' expect only one column but %d." % ('do_get_with_limit', len(result))
        logger.error('%s  \n\t\t sql: %s \n\t\t args: %s' % (msg, sql, args))
        raise MultiColumnsError(msg)
    return None


@with_connection
def do_query(sql: str, *args):
    """
    Execute select SQL and return list results(dict).
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
    """
    global _DB_CTX
    cursor = None
    sql = _before_execute('do_query', sql.strip(), *args)
    try:
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        results = cursor.fetchall()
        if results and cursor.description:
            names = [x[0] for x in cursor.description]
            return [Dict(names, x) for x in results]
        else:
            return results
    finally:
        if cursor:
            cursor.close()


@with_connection
def do_query_one(sql: str, *args):
    """
    execute select SQL and return unique result(dict). Automatically add 'limit ?' behind the sql statement if not.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
    """
    do_sql_log('do_query_one', sql, *args)
    sql, args = _limit_one_sql_args(sql, *args)
    return do_query_one_limit(sql, *args)


@with_connection
def do_query_one_limit(sql: str, *args):
    """
    execute select SQL and return unique result(dict), SQL contain 'limit'.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
    """
    global _DB_CTX
    cursor = None
    sql = _before_execute('do_query_one_with_limit', sql.strip(), *args)
    try:
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        result = cursor.fetchone()
        if result and cursor.description:
            names = [x[0] for x in cursor.description]
            return Dict(names, result)
        return result
    finally:
        if cursor:
            cursor.close()


@with_connection
def do_select(sql: str, *args):
    """
    execute select SQL and return unique result or list results(tuple).
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
    """
    global _DB_CTX
    cursor = None
    sql = _before_execute('do_select', sql.strip(), *args)
    try:
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        return cursor.fetchall()
    finally:
        if cursor:
            cursor.close()


def do_select_one(sql: str, *args):
    """
    Execute select SQL and return unique result(tuple). Automatically add 'limit ?' behind the sql statement if not.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
    """
    do_sql_log('do_select_one', sql, *args)
    sql, args = _limit_one_sql_args(sql, *args)
    return do_select_one_limit(sql, *args)


@with_connection
def do_select_one_limit(sql: str, *args):
    """
    Execute select SQL and return unique result(tuple), SQL contain 'limit'.
    sql: SELECT * FROM user WHERE name=? and age=? limit 1  -->  args: ('张三', 20)
    """
    global _DB_CTX
    cursor = None
    sql = _before_execute('do_select_one_with_limit', sql.strip(), *args)
    try:
        cursor = _DB_CTX.cursor()
        cursor.execute(sql, args)
        return cursor.fetchone()
    finally:
        if cursor:
            cursor.close()


def do_query_page(sql: str, page_num=1, page_size=10, *args):
    """
    Execute select SQL and return list results(dict).
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
    """
    do_page_log('do_query_page', sql, page_num, page_size, args)
    sql, args = _page_sql_args(sql, page_num, page_size, *args)
    return do_query(sql, *args)


def do_select_page(sql: str, page_num=1, page_size=10, *args):
    """
    Execute select SQL and return list results(dict).
    sql: SELECT * FROM user WHERE name=? and age=?  -->  args: ('张三', 20)
    """
    do_page_log('do_select_page', sql, page_num, page_size, args)
    sql, args = _page_sql_args(sql, page_num, page_size, *args)
    return do_select(sql, *args)


def get_connection():
    global _DB_CTX
    if _DB_CTX.is_not_init():
        _DB_CTX.init()
    return _DB_CTX.connection


def prepare(prepared=True):
    global _DB_CTX
    _DB_CTX.prepared = prepared


def _before_execute(function: str, sql: str, *args):
    global _SHOW_SQL
    if _SHOW_SQL:
        logger.info("Exec func 'mysqlx.db.%s' \n\t\tSQL: %s \n\t\tARGS: %s" % (function, sql, args))
    return sql.replace('?', '%s')


def _insert_sql_args(table: str, **kwargs):
    cols, args = zip(*kwargs.items())
    sql = _create_insert_sql(table, cols)
    return sql, args


def _batch_insert_sql_args(table: str, *args):
    args = get_batch_args(*args)
    args = [zip(*arg.items()) for arg in args]  # [(cols, args)]
    cols, args = zip(*args)  # (cols), (args)
    sql = _create_insert_sql(table, cols[0])
    return sql, args


def _batch_named_sql_args(sql: str, *args):
    args = get_batch_args(*args)
    args = [get_named_args(sql, **arg) for arg in args]
    sql = get_named_sql(sql)
    return sql, args


@functools.lru_cache(maxsize=128)
def _create_insert_sql(table: str, cols: Sequence[str]):
    columns, placeholders = zip(*[('`{}`'.format(col), '?') for col in cols])
    return 'INSERT INTO `{}` ({}) VALUES ({})'.format(table, ','.join(columns), ','.join(placeholders))


def _page_sql_args(sql: str, page_num=1, page_size=10, *args):
    start = (page_num - 1) * page_size
    args = [*args, start, page_size]
    if _require_limit(sql):
        sql = '{} limit ?,?'.format(sql)
    return sql, args


def _limit_one_sql_args(sql: str, *args):
    if _require_limit(sql):
        return '{} limit ?'.format(sql), [*args, LIMIT_1]

    return sql, args


def _require_limit(sql: str):
    lower_sql = sql.lower()
    if 'limit' not in lower_sql:
        return True

    idx = lower_sql.rindex('limit')
    if idx > 0 and ')' in lower_sql[idx:]:
        return True

    return False



from logging import basicConfig, INFO, getLogger

logger = getLogger(__name__)
basicConfig(level=INFO, format='[%(levelname)s]: %(asctime)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


def sql_log(function: str, sql: str, *args, **kwargs):
    logger.debug("Exec func 'mysqlx.db.%s' \n\t\tsql: %s \n\t\targs: %s \n\t\tkwargs: %s" % (function, sql.strip(), args, kwargs))


def page_log(function: str, sql: str, page_num, page_size, *args, **kwargs):
    logger.debug("Exec func 'mysqlx.db.%s', page_num: %d, page_size: %d \n\t\tsql: %s \n\t\targs: %s \n\t\tkwargs: %s" % (
        function, page_num, page_size, sql.strip(), args, kwargs))


def insert_log(function: str, table: str, **kwargs):
    logger.debug("Exec func 'mysqlx.db.%s' \n\t\t Table: '%s', kwargs: %s" % (function, table, kwargs))


def do_sql_log(function: str, sql: str, *args):
    logger.debug("Exec func 'mysqlx.db.%s' \n\t\t sql: %s \n\t\t args: %s" % (function, sql, args))


def do_page_log(function: str, sql: str, page_num, page_size, *args):
    logger.debug(
        "Exec func 'mysqlx.db.%s', page_num: %d, page_size: %d \n\t\t sql: %s \n\t\t args: %s" % (function, page_num, page_size, sql.strip(), args))


def sql_id_log(function: str, sql_id: str, *args, **kwargs):
    logger.debug("Exec func 'mysqlx.dbx.%s', sql_id: %s, args: %s, kwargs: %s" % (function, sql_id.strip(), args, kwargs))


def page_sql_id_log(function: str, sql_id: str, page_num, page_size, *args, **kwargs):
    logger.debug("Exec func 'mysqlx.dbx.%s', page_num: %d, page_size: %d, sql_id: %s, args: %s, kwargs: %s" % (
        function, page_num, page_size, sql_id.strip(), args, kwargs))


def orm_by_page_log(function, page_num, page_size, class_name, where, *args, **kwargs):
    logger.debug("Exec func 'mysqlx.orm.Model.%s', page_num: %d, page_size: %d \n\t\t Class: '%s', where: %s, args: %s, kwargs: %s" % (
        function, page_num, page_size, class_name, where, args, kwargs))


def orm_page_log(function, page_num, page_size, class_name, *fields, **kwargs):
    logger.debug("Exec func 'mysqlx.orm.Model.%s', page_num: %d, page_size: %d \n\t\t Class: '%s', fields: %s, kwargs: %s" % (
        function, page_num, page_size, class_name, fields, kwargs))


def orm_insert_log(function, class_name, **kwargs):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', kwargs: %s" % (function, class_name, kwargs))


def orm_delete_by_id_log(function, class_name, _id, update_by):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', id: %d, update_by: %s" % (function, class_name, _id, update_by))


def orm_by_log(function, class_name, where, *args, **kwargs):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', where: %s, args: %s, kwargs: %s" % (function, class_name, where, args, kwargs))


def orm_inst_log(function, class_name):
    logger.debug("Exec func 'mysqlx.orm.Model.%s', Class: '%s'" % (function, class_name))


def orm_logical_delete_by_ids_log(function, class_name, ids, update_by, batch_size):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', ids: %s, update_by: %s, batch_size: %s" % (
    function, class_name, ids, update_by, batch_size))


def orm_count_log(function, class_name, **kwargs):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', kwargs: %s" % (function, class_name, kwargs))


def orm_find_log(function, class_name, *fields, **kwargs):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', fields: %s, kwargs: %s" % (function, class_name, fields, kwargs))


def orm_find_by_id_log(function, class_name, _id, *fields):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', id: %d, fields: %s" % (function, class_name, _id, fields))


def orm_find_by_ids_log(function, class_name, ids, *fields):
    logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', ids: %s, fields: %s" % (function, class_name, ids, fields))



LIMIT_1 = 1

NO_LIMIT = 0

NAMED_REGEX = r':[\w|\d]*'

DYNAMIC_REGEX = '{%|{{|}}|%}'

MAPPER_PATH = "mapper_path"

PK_SQL = 'SELECT LAST_INSERT_ID()'

DEFAULT_PK_FIELD = 'id'

SYMBOLS = ['=', '>', '<']

BETWEEN, LIKE, IN = 'between', 'like', 'in'

PK, TABLE, UPDATE_BY, UPDATE_TIME, DEL_FLAG, PK_STRATEGY = '__pk__', '__table__', '__update_by__', '__update_time__', '__del_flag__', '__pk_strategy__'

COLUMN_SQL = '''SELECT GROUP_CONCAT(CONCAT("`",column_name,"`") SEPARATOR ",") 
                 FROM information_schema.columns WHERE table_schema = (SELECT DATABASE()) AND table_name = ? limit ?'''



import sys
from datetime import datetime
from .snowflake import get_id
from enum import Enum, IntEnum
from functools import lru_cache
from .support import DBError, simple_sql
from typing import Sequence, Union, List, Tuple
from .constant import LIMIT_1, NO_LIMIT, DEFAULT_PK_FIELD, SYMBOLS, BETWEEN, LIKE, IN, PK, TABLE, UPDATE_BY, UPDATE_TIME, DEL_FLAG, PK_STRATEGY, \
    COLUMN_SQL
from .db import do_get_limit, do_query, do_query_one_limit, do_execute, insert, save, do_select, do_select_one_limit, transaction, batch_insert, \
    do_query_page, do_select_page, do_get
from .log_support import logger, orm_page_log, orm_insert_log, orm_by_log, orm_delete_by_id_log, orm_by_page_log, orm_inst_log, orm_find_by_id_log, \
    orm_logical_delete_by_ids_log, orm_count_log, orm_find_log, orm_find_by_ids_log


class DelFlag(IntEnum):
    UN_DELETE = 0
    DELETED = 1


class PkStrategy(Enum):
    """
    SNOWFLAKE: 由Snowflake算法生成主键
    DB_AUTO_INCREMENT: 由数据库的AUTO_INCREMENT自动生成主键

    在Windows上，使用Snowflake可能会报下列错误，这是因为Snowflake生成的id是15位的数字，而Windows上C语言的long类型是32位的
    OverflowError: Python int too large to convert to C long

    如果用的是mysql.connector，且在Windows上开发测试，可以就在初始化数据库的时候加上参数'use_pure'为True用纯python的connect; 在linux是部署生成环境时去掉'use_pure'用
    C语言写的connect, 以提高性能.
    """
    SNOWFLAKE = 'snowflake'
    DB_AUTO_INCREMENT = 'db_auto_increment'


class Model:
    """
    Create a class extends Model:

    class User(Model):
        __pk__ = 'id'
        __table__ = 'user'
        __update_by__ = 'update_by'
        __update_time__ = 'update_time'
        __del_flag__ = 'del_flag'

        def __init__(self, id: int = None, name: str = None, age: int = None, update_by: int = None, update_time: datetime = None, del_flag: int = None):
            self.id = id
            self.update_by = update_by
            self.update_time = update_time
            self.del_flag = del_flag
            self.name = name
            self.age = age

    then you can use like follow:
        db.init_db(user='xxx', password='xxx', database='xxx', host='xxx', ...)  # or dbx.init_db(...) init db first,
        user = User(name='张三', age=55)
        effect_rowcount = user.persist()
        id = user.inst_save()
    """

    def __str__(self):
        return str({k: v for k, v in self.__dict__.items() if not k.startswith("__")})

    def __getattr__(self, name):
        if PK == name:
            return self._get_pk()
        elif TABLE == name:
            return self._get_table()
        elif UPDATE_BY == name:
            return self._get_update_by_field()
        elif UPDATE_TIME == name:
            return self._get_update_time_field()
        else:
            return None

    def persist(self):
        """
        user = User(name='张三', age=55)
        effect_rowcount = user.persist()
        :return: effect rowcount
        """
        orm_inst_log('persist', self.__class__.__name__)
        kv = {k: v for k, v in self.__dict__.items() if v is not None}
        return self.insert(**kv)

    def inst_save(self):
        """
        user = User(name='张三', age=55)
        id = user.save()
        :return: Primary key
        """
        orm_inst_log('inst_save', self.__class__.__name__)
        kv = {k: v for k, v in self.__dict__.items() if v is not None}
        pk = self._get_pk()
        _id = self.save(**kv)
        if pk not in kv:
            self.__dict__.update({pk: _id})
        return _id

    def update(self):
        """
        user = User(id=1, name='李四', age=66)
        rowcount = user.update()
        :return: Effect rowcount
        """
        orm_inst_log('update', self.__class__.__name__)
        pk, table = self._get_pk_and_table()
        kv = {k: v for k, v in self.__dict__.items() if v is not None}
        if pk not in kv:
            raise KeyError("Not primary key.")

        update_kv = {k: v for k, v in kv.items() if k != pk}
        if update_kv:
            return self.update_by_id(kv[pk], **update_kv)
        else:
            logger.warning("Exec func 'mysqlx.orm.Model.%s' not set fields, Class: '%s:'\n\t\t   %s" % ('update', self.__class__.__name__, self))
            return 0

    def load(self, *fields):
        """
        Return new object from database and update itself.
        :param fields: Default select all fields if not set. like: ('id', 'name', 'age')
        user = User(id=1)
        user2 = user.load()
        """
        logger.debug("Exec func 'mysqlx.orm.Model.%s', Class: '%s', fields: %s" % ('load', self.__class__.__name__, fields))
        pk = self._get_pk()
        kv = self.__dict__
        _id = kv.get(pk)
        if _id is not None:
            if not fields:
                fields, _ = zip(*kv.items())
            m = self.query_by_id(_id, *fields)
            if m:
                self.__dict__.update(m)
                return self
            else:
                msg = "Exec func 'mysqlx.orm.Model.%s' load none, Class: '%s', %s=%d." % ('load', self.__class__.__name__, self._get_pk(), _id)
                logger.error(msg)
                raise DBError(msg)
        else:
            raise KeyError("Not primary key.")

    def logical_delete(self):
        """
        Logic delete only update the del flag
        user = User(id=1)
        rowcount = user.logical_delete()
        """
        orm_inst_log('logical_delete', self.__class__.__name__)
        pk = self._get_pk()
        _id = self.__dict__.get(pk)
        update_by = self.__dict__.get(self._get_update_by_field())
        if _id is None:
            raise KeyError("Not primary key.")

        return self.logical_delete_by_id(_id, update_by)

    def un_logical_delete(self):
        """
        Logic un delete only update the del flag
        user = User(id=1)
        rowcount = user.un_logical_delete()
        """
        orm_inst_log('un_logical_delete', self.__class__.__name__)
        pk = self._get_pk()
        _id = self.__dict__.get(pk)
        update_by = self.__dict__.get(self._get_update_by_field())
        if _id is None:
            raise KeyError("Not primary key.")
        return self.un_logical_delete_by_id(_id, update_by)

    def delete(self):
        """
        Physical delete
        user = User(id=1)
        rowcount = user.delete()
        """
        orm_inst_log('delete', self.__class__.__name__)
        pk = self._get_pk()
        _id = self.__dict__.get(pk)
        if _id is None:
            raise KeyError("Not primary key.")
        return self.delete_by_id(_id)

    # ----------------------------------------------------------Class method------------------------------------------------------------------
    @classmethod
    def insert(cls, **kwargs):
        """
        rowcount = User.insert(name='张三', age=20)
        return: Effect rowcount
        """
        orm_insert_log('insert', cls.__name__, **kwargs)
        pk, table = cls._get_pk_and_table()
        pk_strategy = cls._get_pk_strategy()
        if pk_strategy == PkStrategy.SNOWFLAKE and pk not in kwargs:
            kwargs[pk] = get_id()
        return insert(table, **kwargs)

    @classmethod
    def save(cls, **kwargs):
        """
        id = User.save(name='张三', age=20)
        :return: Primary key
        """
        orm_insert_log('save', cls.__name__, **kwargs)
        pk, table = cls._get_pk_and_table()
        pk_strategy = cls._get_pk_strategy()
        if pk_strategy == PkStrategy.SNOWFLAKE:
            if pk in kwargs:
                _id = kwargs[pk]
            else:
                _id = get_id()
                kwargs[pk] = _id
            insert(table, **kwargs)
        else:
            _id = save(table, **kwargs)
        return _id

    @classmethod
    def create(cls, **kwargs):
        """
        user = User.create(name='张三', age=20)
        :return: Instance object
        """
        orm_insert_log('create', cls.__name__, **kwargs)
        pk = cls._get_pk()
        _id = cls.save(**kwargs)
        if pk not in kwargs:
            kwargs[pk] = _id
        return cls._dict2obj(kwargs)

    @classmethod
    def update_by_id(cls, _id: Union[int, str], **kwargs):
        """
        rowcount = User.update_by_id(id=1, name='王五')
        return: Effect rowcount
        """
        logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', id: %d, kwargs: %s" % ('update_by_id', cls.__name__, _id, kwargs))
        assert kwargs, 'Must set update kv'
        pk = cls._get_pk()
        where = '`%s`=?' % pk
        cols, args = zip(*kwargs.items())
        sql, update_time_arg = cls._update_sql(where, *cols)
        if update_time_arg:
            args = [*args, update_time_arg]
        return do_execute(sql, *args, _id, LIMIT_1)

    @classmethod
    def logical_delete_by_id(cls, _id: Union[int, str], update_by: Union[int, str] = None):
        """
        Logic delete only update the del flag
        rowcount = User.delete_by_id(id=1, update_by=100)
        return: Effect rowcount
        """
        orm_delete_by_id_log('logical_delete_by_id', cls.__name__, _id, update_by)
        return cls._logical_delete_by_id_op(_id, update_by, DelFlag.DELETED)

    @classmethod
    def un_logical_delete_by_id(cls, _id: Union[int, str], update_by: Union[int, str] = None):
        """
        Logic delete only update the del flag
        rowcount = User.un_logical_delete_by_id(id=1, update_by=100)
        return: Effect rowcount
        """
        orm_delete_by_id_log('un_logical_delete_by_id', cls.__name__, _id, update_by)
        return cls._logical_delete_by_id_op(_id, update_by, DelFlag.UN_DELETE)

    @classmethod
    def logical_delete_by_ids(cls, ids: Union[Sequence[int], Sequence[str]], update_by: Union[int, str] = None, batch_size=128):
        """
        Logic delete only update the del flag
        rowcount = User.logical_delete_by_ids(id=[1,2], update_by=100)
        return: Effect rowcount
        """
        orm_logical_delete_by_ids_log('logical_delete_by_ids', cls.__name__, ids, update_by, batch_size)
        return cls._logical_delete_by_ids_op(ids, update_by=update_by, batch_size=batch_size, del_status=DelFlag.DELETED)

    @classmethod
    def un_logical_delete_by_ids(cls, ids: Union[Sequence[int], Sequence[str]], update_by: Union[int, str] = None, batch_size=128):
        """
        Logic delete only update the del flag
        rowcount = User.un_logical_delete_by_ids(id=[1,2], update_by=100)
        return: Effect rowcount
        """
        orm_logical_delete_by_ids_log('un_logical_delete_by_ids', cls.__name__, ids, update_by, batch_size)
        return cls._logical_delete_by_ids_op(ids, update_by=update_by, batch_size=batch_size, del_status=DelFlag.UN_DELETE)

    @classmethod
    def delete_by(cls, where: str, *args, **kwargs):
        """
        Physical delete
        rowcount = User.delete_by('where name=? and age=?', '张三', 55)
        return: Effect rowcount
        """
        orm_by_log(sys._getframe().f_code.co_name, cls.__name__, where, *args, **kwargs)
        assert where.lower().startswith('where'), "Must start with 'WHERE' in the where parameter."
        table = cls._get_table()
        sql = 'DELETE FROM `%s` %s' % (table, where)
        sql, args = simple_sql(sql, *args, **kwargs)
        return do_execute(sql, *args)

    @classmethod
    def delete_by_id(cls, _id: Union[int, str]):
        """
        Physical delete
        rowcount = User.delete_by_id(id=1)
        return: Effect rowcount
        """
        logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', id: %d" % ('delete_by_id', cls.__name__, _id))
        pk, table = cls._get_pk_and_table()
        sql = 'DELETE FROM `%s` WHERE `%s`=? limit ?' % (table, pk)
        return do_execute(sql, _id, LIMIT_1)

    @classmethod
    def delete_by_ids(cls, ids: Union[Sequence[int], Sequence[str]], batch_size=128):
        """
        Physical delete
        rowcount = User.delete_by_ids(id=[1,2])
        return: Effect rowcount
        """
        logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', ids: %s, batch_size: %s" % ('delete_by_ids', cls.__name__, ids, batch_size))
        ids_size = len(ids)
        assert ids_size > 0, 'ids must not be empty.'

        if ids_size == 1:
            return cls.delete_by_id(ids[0])
        elif ids_size <= batch_size:
            return cls._delete_by_ids(ids)
        else:
            split_ids = _split_ids(ids, batch_size)
            with transaction():
                results = list(map(cls._delete_by_ids, split_ids))
            return sum(results)

    @classmethod
    def batch_insert(cls, *args):
        """
        Batch insert
        rowcount = User.batch_insert([{'name': '张三', 'age': 55},{'name': '李四', 'age': 66}])
        :param args: All number must have same key.
        :return: Effect rowcount
        """
        logger.debug("Exec func 'mysqlx.orm.Model.%s' \n\t\t Class: '%s', args: %s" % ('batch_insert', cls.__name__, args))
        assert len(args) > 0, 'args must not be empty.'
        pk, table = cls._get_pk_and_table()
        pk_strategy = cls._get_pk_strategy()
        if pk_strategy == PkStrategy.SNOWFLAKE:
            for arg in args:
                if pk not in arg:
                    arg[pk] = get_id()

        return batch_insert(table, *args)

    # ------------------------------------------------Class query method--------------------------------------------------------
    @classmethod
    def count(cls, **kwargs):
        """
        count = User.count(name='张三', age=55)
        """
        orm_count_log('count', cls.__name__, **kwargs)
        table = cls._get_table()
        where, args, _ = _get_where_arg_limit(**kwargs)
        fields = 'count(1)'
        sql = _select_sql(table, where, LIMIT_1, fields)
        return do_get_limit(sql, *args, LIMIT_1)

    @classmethod
    def count_by(cls, where: str, *args, **kwargs):
        """
        Automatically add 'limit ?' where if not.
        count = User.count_by('where name=?', '李四')
        """
        orm_by_log(sys._getframe().f_code.co_name, cls.__name__, where, *args, **kwargs)
        assert where.lower().startswith('where'), "Must start with 'where' in the where parameter."
        table = cls._get_table()
        sql = "SELECT count(1) FROM `{}` {}".format(table, where)
        sql, args = simple_sql(sql, *args, **kwargs)
        return do_get(sql, *args)

    @classmethod
    def exists(cls, **kwargs):
        orm_count_log('exists', cls.__name__, **kwargs)
        return cls.count(**kwargs) > 0

    @classmethod
    def exists_by(cls, where: str, *args, **kwargs):
        orm_by_log(sys._getframe().f_code.co_name, cls.__name__, where, *args, **kwargs)
        return cls.count_by(where, *args, **kwargs) > 0

    @classmethod
    def find(cls, *fields, **kwargs):
        """
        Return list(object) or empty list if no result.
        users = User.find('id', 'name', 'age', name='张三', age=55)
        :param fields: Default select all fields if not set
        """
        orm_find_log('find', cls.__name__, *fields, **kwargs)
        return [cls._dict2obj(d) for d in cls.query(*fields, **kwargs)]

    @classmethod
    def find_one(cls, *fields, **kwargs):
        """
        Return unique result(object) or None if no result.
        user = User.find_one('id', 'name', 'age', name='张三', age=55)
        :param fields: Default select all fields if not set
        """
        orm_find_log('find_one', cls.__name__, *fields, **kwargs)
        result = cls.query_one(*fields, **kwargs)
        return cls._dict2obj(result) if result else None

    @classmethod
    def find_by(cls, where: str, *args, **kwargs):
        """
        Return list(dict) or empty list if no result.
        rows = User.find_by('where name=?', '李四')
        """
        orm_by_log(sys._getframe().f_code.co_name, cls.__name__, where, *args, **kwargs)
        return [cls._dict2obj(d) for d in cls.query_by(where, *args, **kwargs)]

    @classmethod
    def find_page(cls, page_num=1, page_size=10, *fields, **kwargs):
        """
        Return list(object) or empty list if no result.
        users = User.find_page(1, 10, 'name', 'age', name='张三', age=55)
        :param page_num: page number
        :param page_size: page size
        :param fields: Default select all fields if not set
        """
        orm_page_log('find_page', page_num, page_size, cls.__name__, *fields, **kwargs)
        result = cls.query_page(page_num, page_size, *fields, **kwargs)
        return [cls._dict2obj(d) for d in result]

    @classmethod
    def find_page_by(cls, page_num: int, page_size: int, where: str, *args, **kwargs):
        """
        Return list(dict) or empty list if no result. Automatically add 'limit ?,?' after where if not.
        rows = User.find_by_page(1, 10, 'where name=?', '李四')
        """
        orm_page_log(sys._getframe().f_code.co_name, page_num, page_size, cls.__name__, where, *args, **kwargs)
        return [cls._dict2obj(d) for d in cls.query_page_by(page_num, page_size, where, *args, **kwargs)]

    @classmethod
    def find_by_id(cls, _id: Union[int, str], *fields):
        """
        Return one class object or None if no result.
        user = User.find_by_id(1, 'id', 'name', 'age')
        :param _id: pk
        :param fields: Default select all fields if not set
        """
        orm_find_by_id_log('find_by_id', cls.__name__, _id, *fields)
        result = cls.query_by_id(_id, *fields)
        return cls._dict2obj(result) if result else None

    @classmethod
    def find_by_ids(cls, ids: Union[Sequence[int], Sequence[str]], *fields):
        """
        Return list(class object) or empty list if no result.
        users = User.find_by_ids([1,2], 'id', 'name', 'age')
        :param ids: List of pk
        :param fields: Default select all fields if not set
        """
        orm_find_by_ids_log('find_by_ids', cls.__name__, ids, *fields)
        return [cls._dict2obj(d) for d in cls.query_by_ids(ids, *fields)]

    @classmethod
    def query(cls, *fields, **kwargs):
        """
        Return list(dict) or empty list if no result.
        users = User.query('id', 'name', 'age', name='张三', age=55)
        :param fields: Default select all fields if not set
        """
        orm_find_log('query', cls.__name__, *fields, **kwargs)
        where, args, limit = _get_where_arg_limit(**kwargs)
        table = cls._get_table()
        sql = _select_sql(table, where, limit, *fields)
        if limit:
            if isinstance(limit, int):
                args = [*args, limit]
            else:
                args = [*args, *limit]
        return do_query(sql, *args)

    @classmethod
    def query_one(cls, *fields, **kwargs):
        """
        Return unique result(dict) or None if no result.
        users = User.query_one('id', 'name', 'age', name='张三', age=55)
        :param fields: Default select all fields if not set
        """
        orm_find_log('query_one', cls.__name__, *fields, **kwargs)
        where, args, _ = _get_where_arg_limit(**kwargs)
        table = cls._get_table()
        sql = _select_sql(table, where, LIMIT_1, *fields)
        return do_query_one_limit(sql, *args, LIMIT_1)

    @classmethod
    def query_by(cls, where: str, *args, **kwargs):
        """
        Return list(dict) or empty list if no result.
        rows = User.query_by('where name=?', '李四')
        """
        orm_by_log(sys._getframe().f_code.co_name, cls.__name__, where, *args, **kwargs)
        sql = cls._where_sql(where)
        sql, args = simple_sql(sql, *args, **kwargs)
        return do_query(sql, *args)

    @classmethod
    def query_page(cls, page_num=1, page_size=10, *fields, **kwargs):
        """
        Return list(dict) or empty list if no result.
        users = User.query_page(1, 10, 'id', 'name', 'age', name='张三', age=55)
        :param page_num: page number
        :param page_size: page size
        :param fields: Default select all fields if not set
        """
        orm_page_log('query_page', page_num, page_size, cls.__name__, *fields, **kwargs)
        table = cls._get_table()
        where, args, _ = _get_where_arg_limit(**kwargs)
        sql = _select_sql(table, where, NO_LIMIT, *fields)
        return do_query_page(sql, page_num, page_size, *args)

    @classmethod
    def query_page_by(cls, page_num: int, page_size: int, where: str, *args, **kwargs):
        """
        Return list(dict) or empty list if no result. Automatically add 'limit ?,?' after where if not.
        rows = User.query_by_page(1, 10, 'where name=?', '李四')
        """
        orm_by_page_log(sys._getframe().f_code.co_name, page_num, page_size, cls.__name__, where, *args, **kwargs)
        sql = cls._where_sql(where)
        sql, args = simple_sql(sql, *args, **kwargs)
        return do_query_page(sql, page_num, page_size, *args)

    @classmethod
    def query_by_id(cls, _id: Union[int, str], *fields):
        """
        Return one row(dict) or None if no result.
        user = User.query_by_id(1, 'id', 'name', 'age')
        :param _id: pk
        :param fields: Default select all fields if not set
        """
        orm_find_by_id_log('query_by_id', cls.__name__, _id, *fields)
        pk, table = cls._get_pk_and_table()
        where = 'WHERE `{}`=?'.format(pk)
        sql = _select_sql(table, where, LIMIT_1, *fields)
        return do_query_one_limit(sql, _id, LIMIT_1)

    @classmethod
    def query_by_ids(cls, ids: Union[Sequence[int], Sequence[str]], *fields):
        """
        Return list(dict) or empty list if no result.
        users = User.query_by_ids([1,2], 'id', 'name', 'age')
        :param ids: List of pk
        :param fields: Default select all fields if not set
        """
        orm_find_by_ids_log('query_by_ids', cls.__name__, ids, *fields)
        ids_size = len(ids)
        assert ids_size > 0, 'ids must not be empty.'

        pk, table = cls._get_pk_and_table()
        where = 'WHERE `{}` in ({})'.format(pk, ','.join(['?' for _ in range(ids_size)]))
        sql = _select_sql(table, where, ids_size, *fields)
        return do_query(sql, *ids, ids_size)

    @classmethod
    def select(cls, *fields, **kwargs):
        """
        Return list(dict) or empty list if no result.
        rows = User.select('id', 'name', 'age', name='张三', age=55)
        :param fields: Default select all fields if not set
        """
        orm_find_log('select', cls.__name__, *fields, **kwargs)
        where, args, limit = _get_where_arg_limit(**kwargs)
        table = cls._get_table()
        sql = _select_sql(table, where, limit, *fields)
        if limit:
            if isinstance(limit, int):
                args = [*args, limit]
            else:
                args = [*args, *limit]
        return do_select(sql, *args)

    @classmethod
    def select_one(cls, *fields, **kwargs):
        """
        Return unique result(tuple) or None if no result.
        row = User.select_one('id', 'name', 'age', name='张三', age=55)
        :param fields: Default select all fields if not set
        """
        orm_find_log('select_one', cls.__name__, *fields, **kwargs)
        where, args, _ = _get_where_arg_limit(**kwargs)
        table = cls._get_table()
        sql = _select_sql(table, where, LIMIT_1, *fields)
        return do_select_one_limit(sql, *args, LIMIT_1)

    @classmethod
    def select_by(cls, where: str, *args, **kwargs):
        """
        Return list(dict) or empty list if no result.
        rows = User.select_by('where name=?', '李四')
        """
        orm_by_log(sys._getframe().f_code.co_name, cls.__name__, where, *args, **kwargs)
        sql = cls._where_sql(where)
        sql, args = simple_sql(sql, *args, **kwargs)
        return do_select(sql, *args)

    @classmethod
    def select_page(cls, page_num=1, page_size=10, *fields, **kwargs):
        """
        Return list(dict) or empty list if no result.
        rows = User.select_page('id', 'name', 'age', name='张三', age=55)
        :param page_num: page number
        :param page_size: page size
        :param fields: Default select all fields if not set
        """
        orm_page_log('select_page', page_num, page_size, cls.__name__, *fields, **kwargs)
        table = cls._get_table()
        where, args, _ = _get_where_arg_limit(**kwargs)
        sql = _select_sql(table, where, NO_LIMIT, *fields)
        return do_select_page(sql, page_num, page_size, *args)

    @classmethod
    def select_page_by(cls, page_num: int, page_size: int, where: str, *args, **kwargs):
        """
        Return list(dict) or empty list if no result. Automatically add 'limit ?,?' after where if not.
        rows = User.select_by_page(1, 10, 'where name=?', '李四')
        """
        orm_by_page_log(sys._getframe().f_code.co_name, page_num, page_size, cls.__name__, where, *args, **kwargs)
        sql = cls._where_sql(where)
        sql, args = simple_sql(sql, *args, **kwargs)
        return do_select_page(sql, page_num, page_size, *args)

    @classmethod
    def select_by_id(cls, _id: Union[int, str], *fields):
        """
        Return one row(dict) or None if no result.
        row = User.select_by_id(1, 'id', 'name', 'age')
        :param _id: pk
        :param fields: Default select all fields if not set
        """
        orm_find_by_id_log('select_by_id', cls.__name__, _id, *fields)
        pk, table = cls._get_pk_and_table()
        where = 'WHERE `{}`=?'.format(pk)
        sql = _select_sql(table, where, LIMIT_1, *fields)
        return do_select_one_limit(sql, _id, LIMIT_1)

    @classmethod
    def select_by_ids(cls, ids: Union[Sequence[int], Sequence[str]], *fields):
        """
        Return list(dict) or empty list if no result.
        rows = User.select_by_ids([1,2], 'id', 'name', 'age')
        :param ids: List of pk
        :param fields: Default select all fields if not set
        """
        orm_find_by_ids_log('select_by_ids', cls.__name__, ids, *fields)
        ids_size = len(ids)
        assert ids_size > 0, 'ids must not be empty.'

        pk, table = cls._get_pk_and_table()
        where = 'WHERE `{}` in ({})'.format(pk, ','.join(['?' for _ in range(ids_size)]))
        sql = _select_sql(table, where, ids_size, *fields)
        return do_select(sql, *ids, ids_size)

    # ------------------------------------------------Private class method------------------------------------------------------------------
    @classmethod
    def _logical_delete_by_id_op(cls, _id: Union[int, str], update_by: Union[int, str] = None, del_status=DelFlag.DELETED):
        pk, table = cls._get_pk_and_table()
        del_flag_field = cls._get_del_flag_field()
        update_by_field = cls._get_update_by_field()

        where = '`%s`=?' % pk
        if update_by is not None and update_by_field is not None:
            sql, update_time_arg = cls._update_sql(where, del_flag_field, update_by_field)
            if update_time_arg:
                return do_execute(sql, del_status.value, update_by, update_time_arg, _id, LIMIT_1)
            return do_execute(sql, del_status.value, update_by, _id, LIMIT_1)
        else:
            sql, update_time_arg = cls._update_sql(where, del_flag_field)
            if update_time_arg:
                return do_execute(sql, del_status.value, update_time_arg, _id, LIMIT_1)
            return do_execute(sql, del_status.value, _id, LIMIT_1)

    @classmethod
    def _logical_delete_by_ids_op(cls, ids: Union[Sequence[int], Sequence[str]], update_by: Union[int, str] = None, batch_size=128,
            del_status=DelFlag.DELETED):
        ids_size = len(ids)
        assert ids_size > 0, 'ids must not be empty.'

        if ids_size == 1:
            return cls._logical_delete_by_id_op(ids[0], update_by, del_status)
        elif ids_size <= batch_size:
            return cls._do_logical_delete_by_ids(ids, update_by, del_status)
        else:
            split_ids = _split_ids(ids, batch_size)
            with transaction():
                results = [cls._do_logical_delete_by_ids(ids, update_by, del_status) for ids in split_ids]
            return sum(results)

    @classmethod
    def _do_logical_delete_by_ids(cls, ids: Union[Sequence[int], Sequence[str]], update_by: Union[int, str] = None, del_status=DelFlag.DELETED):
        ids_size = len(ids)
        pk = cls._get_pk()
        del_flag_field = cls._get_del_flag_field()
        update_by_field = cls._get_update_by_field()

        where = '`%s` in (%s)' % (pk, ','.join(['?' for _ in range(ids_size)]))
        if update_by is not None and update_by_field is not None:
            sql, update_time_arg = cls._update_sql(where, del_flag_field, update_by_field)
            if update_time_arg:
                return do_execute(sql, del_status.value, update_by, update_time_arg, *ids, ids_size)
            return do_execute(sql, del_status.value, update_by, *ids, ids_size)
        else:
            sql, update_time_arg = cls._update_sql(where, del_flag_field)
            if update_time_arg:
                return do_execute(sql, del_status.value, update_time_arg, *ids, ids_size)
            return do_execute(sql, del_status.value, *ids, ids_size)

    @classmethod
    def _delete_by_ids(cls, ids: Union[Sequence[int], Sequence[str]]):
        ids_size = len(ids)
        pk, table = cls._get_pk_and_table()
        sql = 'DELETE FROM `{}` WHERE `{}` in ({}) limit ?'.format(table, pk, ','.join(['?' for _ in range(ids_size)]))
        return do_execute(sql, *ids, ids_size)

    @classmethod
    def _get_pk(cls):
        if hasattr(cls, PK):
            return cls.__pk__
        logger.warning("%s not set attribute '%s'" % (cls.__name__, PK))
        return DEFAULT_PK_FIELD

    @classmethod
    def _get_table(cls):
        if hasattr(cls, TABLE):
            return cls.__table__
        logger.warning("%s not set attribute '%s'" % (cls.__name__, TABLE))
        return _get_table_name(cls.__name__)

    @classmethod
    def _get_pk_and_table(cls):
        return cls._get_pk(), cls._get_table()

    @classmethod
    def _get_pk_strategy(cls):
        if hasattr(cls, PK_STRATEGY):
            return cls.__pk_strategy__
        return None

    @classmethod
    def _dict2obj(cls, dictionary):
        m = cls.__new__(cls)
        m.__dict__.update(dictionary)
        return m

    @classmethod
    def _get_update_by_field(cls):
        if hasattr(cls, UPDATE_BY):
            return cls.__update_by__
        return None

    @classmethod
    def _get_update_time_field(cls):
        if hasattr(cls, UPDATE_TIME):
            return cls.__update_time__
        return None

    @classmethod
    def _get_del_flag_field(cls):
        assert hasattr(cls, DEL_FLAG), "%s not set attribute '%s'" % (cls.__name__, DEL_FLAG)
        return cls.__del_flag__

    @classmethod
    def _update_sql(cls, where, *update_fields):
        update_time_arg = None
        table = cls._get_table()
        update_time_field = cls._get_update_time_field()
        if update_time_field is not None and update_time_field not in update_fields:
            update_fields = [*update_fields, update_time_field]
            update_time_arg = datetime.now()

        update_fields = ','.join(['`{}`=?'.format(col) for col in update_fields])
        return 'UPDATE `{}` SET {} WHERE {} limit ?'.format(table, update_fields, where), update_time_arg

    @classmethod
    def _where_sql(cls, where: str):
        low_where = where.lower()
        if low_where.startswith('where'):
            table = cls._get_table()
            return _select_sql(table, where, NO_LIMIT)
        elif low_where.startswith('select'):
            return where
        raise DBError("The where parameter must be a complete SQL statement or conditions start with 'where'")


# ----------------------------------------------------------Private function------------------------------------------------------------------
def _select_sql(table: str, where: str, limit: Union[int, Tuple[int], List[int]], *fields):
    if fields:
        fields = ','.join([col if '(' in col else '`{}`'.format(col) for col in fields])
    else:
        fields = _get_table_columns(table)

    if limit:
        if isinstance(limit, int):
            return 'SELECT {} FROM `{}` {} limit ?'.format(fields, table, where)
        elif (isinstance(limit, Tuple) or isinstance(limit, List)) and len(limit) == 2:
            return 'SELECT {} FROM `{}` {} limit ?,?'.format(fields, table, where)
        else:
            raise ValueError("The type of the parameter 'limit' must be 'int' or tuple, list, and it length is 2.")
    else:
        return 'SELECT {} FROM `{}` {}'.format(fields, table, where)


@lru_cache(maxsize=128)
def _get_table_columns(table: str):
    return do_get_limit(COLUMN_SQL, table, LIMIT_1)


def _delete_by_id_sql(table, pk):
    return 'DELETE FROM `{}` WHERE `{}`=? limit ?'.format(table, pk)


def _get_condition_arg(k: str, v: object):
    if k.endswith("__eq"):
        return "`{}`=?".format(k[:-4]), v
    if k.endswith("__gt"):
        return "`{}`>?".format(k[:-4]), v
    if k.endswith("__lt"):
        return "`{}`<?".format(k[:-4]), v
    if k.endswith("__ge"):
        return "`{}`>=?".format(k[:-4]), v
    if k.endswith("__gte"):
        return "`{}`>=?".format(k[:-5]), v
    if k.endswith("__le"):
        return "`{}`<=?".format(k[:-4]), v
    if k.endswith("__lte"):
        return "`{}`<=?".format(k[:-5]), v
    if k.endswith("__in") and isinstance(v, Sequence) and not isinstance(v, str):
        return "`{}` in ({})".format(k[:-4], ','.join(['?' for _ in v])), v
    if k.endswith("__in"):
        return "`{}` in ({})".format(k[:-4], '?'), v
    if k.endswith("__startswith"):
        return "`{}` like ?".format(k[:-12]), '{}%'.format(v)
    if k.endswith("__endswith"):
        return "`{}` like ?".format(k[:-10]), '%{}'.format(v)
    if k.endswith("__contains"):
        return "`{}` like ?".format(k[:-10]), '%{}%'.format(v)
    if k.endswith("__between") and isinstance(v, Sequence) and 2 == len(v) and not isinstance(v, str):
        return "`{}` between ? and ?".format(k[:-9]), v
    if k.endswith("__between"):
        return ValueError("Must is instance of Sequence with length 2 when use between statement")

    if not isinstance(v, str):
        return "`{}`=?".format(k), v
    lower_v = v.lower()
    if any([symbol in SYMBOLS for symbol in lower_v]):
        return "`{}`{}".format(k, v), None
    elif BETWEEN in lower_v or LIKE in lower_v or IN in lower_v:
        return "`{}` {}".format(k, v), None
    else:
        return "`{}`=?".format(k), v


def _get_where_arg_limit(**kwargs):
    where, args, limit = '', [], 0
    if 'limit' in kwargs:
        limit = kwargs.get('limit')
        del kwargs['limit']

    if kwargs:
        conditions, tmp_args = zip(*[_get_condition_arg(k, v) for k, v in kwargs.items()])
        tmp_args = [arg for arg in tmp_args if arg is not None]

        for arg in tmp_args:
            if arg:
                if isinstance(arg, Sequence) and not isinstance(arg, str):
                    args.extend(arg)
                else:
                    args.append(arg)
        where = 'WHERE {}'.format(' and '.join(conditions))

    return where, args, limit


def _split_ids(ids: Sequence[int], batch_size):
    ids_size = len(ids)
    n = ids_size // batch_size
    n += 0 if ids_size % batch_size == 0 else 1
    return [ids[i:i + batch_size] for i in range(0, ids_size, batch_size)]


def _get_table_name(class_name):
    for i in range(1, len(class_name) - 1)[::-1]:
        if class_name[i].isupper():
            class_name = class_name[:i] + '_' + class_name[i:]
    return class_name.lower()
